
#include <stdio.h>

typedef unsigned int u32;
typedef unsigned char u8;

#define KEYLEN (128)
#define BLOCKLEN (128)
#define ROUNDS (16)

static const u32 sbox0[] = { 12, 15, 7, 10, 14, 13, 11, 0, 2, 6, 3, 1, 9, 4, 5, 8 };
static const u32 sbox1[] = { 7, 2, 14, 9, 3, 11, 0, 4, 12, 13, 1, 10, 6, 15, 8, 5 };
static const u32     o[] = { 7, 6, 2, 1, 5, 0, 3, 4 };
static const u32    pr[] = { 2, 5, 4, 0, 3, 1, 7, 6 };

void lucifer_encrypt(u8 *block, u8 *key)
{
	int i, j, k;
	u32 kc, ks;
	u32 l, h, v;
	u8 *h0, *h1, *tmp;
	  
	h0 = block+0;
	h1 = block+8;
	
	kc = 0;	
  for(i=0; i < ROUNDS; ++i)
  {
  	ks = kc;
  	for(j=0; j < 8; ++j) // byte index
  	{
  		l = (h1[j]>>0)&0x0F;  	
  		h = (h1[j]>>4)&0x0F;
  		
  		v = (key[ks]>>j)&1 ? sbox0[l] | sbox1[h]<<4 : 
  			                   sbox0[h] | sbox1[l]<<4;
			v ^= key[kc];
  	  for(k=0; k < 8; ++k)
  	  {
  	  	int idx = (o[k]+j)%8;
  	  	h0[idx] = (h0[idx] & ~(1<<k)) | ((((h0[idx]>>k) ^ (v>>pr[k]))&1)<<k);  	  	
  	  }
  	  if(j < 7)
  	  	kc = (kc+1)%16;
  	}
  	tmp = h0;
  	h0 = h1;
  	h1 = tmp;
  }
  
  for(i=0; i < 8; ++i)
  {
  	u8 x;
  	x = h0[i];
  	h0[i] = h1[i];
  	h1[i] = x;
  }
}

void lucifer_decrypt(u8 *block, u8 *key)
{
	int i, j, k;
	u32 kc, ks;
	u32 l, h, v;
	u8 *h0, *h1, *tmp;
	  
	h0 = block+0;
	h1 = block+8;
	
	kc = 8;	
  for(i=0; i < ROUNDS; ++i)
  {
  	kc = (kc+1)%16;
  	ks = kc;
  	for(j=0; j < 8; ++j) // byte index
  	{
  		l = (h1[j]>>0)&0x0F;  	
  		h = (h1[j]>>4)&0x0F;
  		
  		v = (key[ks]>>j)&1 ? sbox0[l] + (sbox1[h]<<4) : 
  			                   sbox0[h] + (sbox1[l]<<4);
			v ^= key[kc];
  	  for(k=0; k < 8; ++k)
  	  {
  	  	int idx = (o[k]+j)%8;
  	  	h0[idx] = (h0[idx] & ~(1<<k)) | ((((h0[idx] >> k) ^ (v >> pr[k]))&1) << k);
  	  }
  	  
  	  kc = (kc+1)%16;
  	}
  	tmp = h0;
  	h0 = h1;
  	h1 = tmp;
  }
  
  for(i=0; i < 8; ++i)
  {
  	u8 x;
  	x = h0[i];
  	h0[i] = h1[i];
  	h1[i] = x;
  }
}


int main(int argc, char **argv)
{
	int i;
	u8 key[16], block[16];
	
	memset(key, 0, 16);
	memset(block, 0, 16);
	lucifer_encrypt(block, key);
	for(i=0; i < 16; ++i)
		printf("%02X ", block[i]);
	printf("\n");
	lucifer_decrypt(block, key);
	for(i=0; i < 16; ++i)
		printf("%02X ", block[i]);
	printf("\n");
	return 0;
}

